package ru.dbotthepony.mc.otm.worldgen

import net.minecraft.core.BlockPos
import net.minecraft.util.RandomSource
import net.minecraft.world.level.ChunkPos
import net.minecraft.world.level.WorldGenLevel
import net.minecraft.world.level.chunk.ChunkGenerator
import net.minecraft.world.level.levelgen.feature.FeaturePlaceContext
import net.minecraft.world.level.levelgen.feature.configurations.FeatureConfiguration
import net.minecraft.world.level.levelgen.placement.PlacementContext
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.mc.otm.worldgen.feature.EnhancedFeature
import java.util.*
import kotlin.collections.HashMap

class EnhancedPlacementContext {
	fun interface Placer {
		fun place(context: EnhancedPlacementContext, positions: List<BlockPos>, feature: EnhancedFeature.Configured<*, *>)
	}

	val level: WorldGenLevel
	val generator: ChunkGenerator
	val random: RandomSource
	val origin: ChunkPos
	private val placer: Placer
	val vanillaContext: PlacementContext

	private var parent: EnhancedPlacementContext? = null
	private val variables = HashMap<PlacementVariable<*>, KOptional<*>>()

	private fun <T> recursiveGet(index: PlacementVariable<T>): KOptional<T>? {
		return variables[index] as KOptional<T>? ?: parent?.recursiveGet(index)
	}

	operator fun <T> get(index: PlacementVariable<T>): KOptional<T> {
		return recursiveGet(index) ?: KOptional()
	}

	operator fun <T> set(index: PlacementVariable<T>, value: T) {
		variables[index] = KOptional(value)
	}

	fun <T> remove(index: PlacementVariable<T>): KOptional<T> {
		val old = variables.put(index, KOptional<T>())
		return old as KOptional<T>? ?: KOptional()
	}

	fun remove(index: Collection<PlacementVariable<*>>) {
		index.forEach { remove(it) }
	}

	constructor(
		level: WorldGenLevel,
		generator: ChunkGenerator,
		random: RandomSource,
		origin: ChunkPos,
		placer: Placer,
	) {
		this.level = level
		this.generator = generator
		this.random = random
		this.origin = origin
		this.placer = placer

		vanillaContext = PlacementContext(level, generator, Optional.empty())
	}

	private constructor(parent: EnhancedPlacementContext) {
		this.level = parent.level
		this.generator = parent.generator
		this.random = parent.random
		this.origin = parent.origin
		this.placer = parent.placer
		this.parent = parent
		this.vanillaContext = parent.vanillaContext
	}

	fun push(): EnhancedPlacementContext {
		return EnhancedPlacementContext(this)
	}

	fun place(positions: List<BlockPos>, feature: EnhancedFeature.Configured<*, *>) {
		placer.place(this, positions, feature)
	}

	fun <FC : FeatureConfiguration> vanillaFeatureContext(config: FC, position: BlockPos): FeaturePlaceContext<FC> {
		return FeaturePlaceContext(Optional.empty(), level, generator, random, position, config)
	}
}
